@@ -1,13 +1,5 @@
-import React from 'react';
-import uPlot from 'uplot';
-
 import {
-  ArrayVector,
   DataFrame,
-  DashboardCursorSync,
-  DataHoverPayload,
-  DataHoverEvent,
-  DataHoverClearEvent,
   FALLBACK_COLOR,
   Field,
   FieldColorModeId,
@@ -22,8 +14,14 @@ import {
   getFieldConfigWithMinMax,
   ThresholdsMode,
   TimeRange,
+  cacheFieldDisplayNames,
+  outerJoinDataFrames,
+  ValueMapping,
+  ThresholdsConfig,
 } from '@grafana/data';
-import { maybeSortFrame } from '@grafana/data/src/transformations/transformers/joinDataFrames';
+import { maybeSortFrame, NULL_RETAIN } from '@grafana/data/src/transformations/transformers/joinDataFrames';
+import { applyNullInsertThreshold } from '@grafana/data/src/transformations/transformers/nulls/nullInsertThreshold';
+import { nullToValue } from '@grafana/data/src/transformations/transformers/nulls/nullToValue';
 import {
   VizLegendOptions,
   AxisPlacement,
@@ -32,17 +30,9 @@ import {
   VisibilityMode,
   TimelineValueAlignment,
   HideableFieldConfig,
+  MappingType,
 } from '@grafana/schema';
-import {
-  FIXED_UNIT,
-  SeriesVisibilityChangeMode,
-  UPlotConfigBuilder,
-  UPlotConfigPrepFn,
-  VizLegendItem,
-} from '@grafana/ui';
-import { applyNullInsertThreshold } from '@grafana/ui/src/components/GraphNG/nullInsertThreshold';
-import { nullToValue } from '@grafana/ui/src/components/GraphNG/nullToValue';
-import { PlotTooltipInterpolator } from '@grafana/ui/src/components/uPlot/types';
+import { FIXED_UNIT, UPlotConfigBuilder, UPlotConfigPrepFn, VizLegendItem } from '@grafana/ui';
 import { preparePlotData2, getStackingGroups } from '@grafana/ui/src/components/uPlot/utils';
 
 import { getConfig, TimelineCoreOptions } from './timeline';
@@ -54,13 +44,13 @@ interface UPlotConfigOptions {
   frame: DataFrame;
   theme: GrafanaTheme2;
   mode: TimelineMode;
-  sync?: () => DashboardCursorSync;
   rowHeight?: number;
   colWidth?: number;
   showValue: VisibilityMode;
   alignValue?: TimelineValueAlignment;
   mergeValues?: boolean;
-  getValueColor: (frameIdx: number, fieldIdx: number, value: any) => string;
+  getValueColor: (frameIdx: number, fieldIdx: number, value: unknown) => string;
+  hoverMulti: boolean;
 }
 
 /**
@@ -81,31 +71,22 @@ const defaultConfig: PanelFieldConfig = {
   fillOpacity: 80,
 };
 
-export function mapMouseEventToMode(event: React.MouseEvent): SeriesVisibilityChangeMode {
-  if (event.ctrlKey || event.metaKey || event.shiftKey) {
-    return SeriesVisibilityChangeMode.AppendToSelection;
-  }
-  return SeriesVisibilityChangeMode.ToggleSelection;
-}
-
 export const preparePlotConfigBuilder: UPlotConfigPrepFn<UPlotConfigOptions> = ({
   frame,
   theme,
   timeZones,
   getTimeRange,
   mode,
-  eventBus,
-  sync,
   rowHeight,
   colWidth,
   showValue,
   alignValue,
   mergeValues,
   getValueColor,
+  hoverMulti,
 }) => {
   const builder = new UPlotConfigBuilder(timeZones[0]);
 
-  const xScaleUnit = 'time';
   const xScaleKey = 'x';
 
   const isDiscrete = (field: Field) => {
@@ -113,7 +94,15 @@ export const preparePlotConfigBuilder: UPlotConfigPrepFn<UPlotConfigOptions> = (
     return !(mode && field.display && mode.startsWith('continuous-'));
   };
 
-  const getValueColorFn = (seriesIdx: number, value: any) => {
+  const hasMappedNull = (field: Field) => {
+    return (
+      field.config.mappings?.some(
+        (mapping) => mapping.type === MappingType.SpecialValue && mapping.options.match === 'null'
+      ) || false
+    );
+  };
+
+  const getValueColorFn = (seriesIdx: number, value: unknown) => {
     const field = frame.fields[seriesIdx];
 
     if (
@@ -131,6 +120,7 @@ export const preparePlotConfigBuilder: UPlotConfigPrepFn<UPlotConfigOptions> = (
     mode: mode!,
     numSeries: frame.fields.length - 1,
     isDiscrete: (seriesIdx) => isDiscrete(frame.fields[seriesIdx]),
+    hasMappedNull: (seriesIdx) => hasMappedNull(frame.fields[seriesIdx]),
     mergeValues,
     rowHeight: rowHeight,
     colWidth: colWidth,
@@ -143,56 +133,13 @@ export const preparePlotConfigBuilder: UPlotConfigPrepFn<UPlotConfigOptions> = (
     getTimeRange,
     // hardcoded formatter for state values
     formatValue: (seriesIdx, value) => formattedValueToString(frame.fields[seriesIdx].display!(value)),
-    onHover: (seriesIndex, valueIndex) => {
-      hoveredSeriesIdx = seriesIndex;
-      hoveredDataIdx = valueIndex;
-      shouldChangeHover = true;
-    },
-    onLeave: () => {
-      hoveredSeriesIdx = null;
-      hoveredDataIdx = null;
-      shouldChangeHover = true;
-    },
+    hoverMulti,
   };
 
-  let shouldChangeHover = false;
-  let hoveredSeriesIdx: number | null = null;
-  let hoveredDataIdx: number | null = null;
-
   const coreConfig = getConfig(opts);
-  const payload: DataHoverPayload = {
-    point: {
-      [xScaleUnit]: null,
-      [FIXED_UNIT]: null,
-    },
-    data: frame,
-  };
 
   builder.addHook('init', coreConfig.init);
   builder.addHook('drawClear', coreConfig.drawClear);
-  builder.addHook('setCursor', coreConfig.setCursor);
-
-  // in TooltipPlugin, this gets invoked and the result is bound to a setCursor hook
-  // which fires after the above setCursor hook, so can take advantage of hoveringOver
-  // already set by the above onHover/onLeave callbacks that fire from coreConfig.setCursor
-  const interpolateTooltip: PlotTooltipInterpolator = (
-    updateActiveSeriesIdx,
-    updateActiveDatapointIdx,
-    updateTooltipPosition
-  ) => {
-    if (shouldChangeHover) {
-      if (hoveredSeriesIdx != null) {
-        updateActiveSeriesIdx(hoveredSeriesIdx);
-        updateActiveDatapointIdx(hoveredDataIdx);
-      }
-
-      shouldChangeHover = false;
-    }
-
-    updateTooltipPosition(hoveredSeriesIdx == null);
-  };
-
-  builder.setTooltipInterpolator(interpolateTooltip);
 
   builder.setPrepData((frames) => preparePlotData2(frames[0], getStackingGroups(frames[0])));
 
@@ -270,77 +217,9 @@ export const preparePlotConfigBuilder: UPlotConfigPrepFn<UPlotConfigOptions> = (
     });
   }
 
-  if (sync && sync() !== DashboardCursorSync.Off) {
-    let cursor: Partial<uPlot.Cursor> = {};
-
-    cursor.sync = {
-      key: '__global_',
-      filters: {
-        pub: (type: string, src: uPlot, x: number, y: number, w: number, h: number, dataIdx: number) => {
-          if (sync && sync() === DashboardCursorSync.Off) {
-            return false;
-          }
-          payload.rowIndex = dataIdx;
-          if (x < 0 && y < 0) {
-            payload.point[xScaleUnit] = null;
-            payload.point[FIXED_UNIT] = null;
-            eventBus.publish(new DataHoverClearEvent());
-          } else {
-            payload.point[xScaleUnit] = src.posToVal(x, xScaleKey);
-            payload.point.panelRelY = y > 0 ? y / h : 1; // used for old graph panel to position tooltip
-            payload.down = undefined;
-            eventBus.publish(new DataHoverEvent(payload));
-          }
-          return true;
-        },
-      },
-      scales: [xScaleKey, null],
-    };
-    builder.setSync();
-    builder.setCursor(cursor);
-  }
-
   return builder;
 };
 
-export function getNamesToFieldIndex(frame: DataFrame): Map<string, number> {
-  const names = new Map<string, number>();
-  for (let i = 0; i < frame.fields.length; i++) {
-    names.set(getFieldDisplayName(frame.fields[i], frame), i);
-  }
-  return names;
-}
-
-/**
- * If any sequential duplicate values exist, this will return a new array
- * with the future values set to undefined.
- *
- * in:  1,        1,undefined,        1,2,        2,null,2,3
- * out: 1,undefined,undefined,undefined,2,undefined,null,2,3
- */
-export function unsetSameFutureValues(values: unknown[]): unknown[] | undefined {
-  let prevVal = values[0];
-  let clone: unknown[] | undefined = undefined;
-
-  for (let i = 1; i < values.length; i++) {
-    let value = values[i];
-
-    if (value === null) {
-      prevVal = null;
-    } else {
-      if (value === prevVal) {
-        if (!clone) {
-          clone = [...values];
-        }
-        clone[i] = undefined;
-      } else if (value != null) {
-        prevVal = value;
-      }
-    }
-  }
-  return clone;
-}
-
 function getSpanNulls(field: Field) {
   let spanNulls = field.config.custom?.spanNulls;
 
@@ -371,7 +250,7 @@ export function mergeThresholdValues(field: Field, theme: GrafanaTheme2): Field
     textToColor.set(items[i].label, items[i].color!);
   }
 
-  let input = field.values.toArray();
+  let input = field.values;
   const vals = new Array<String | undefined>(field.values.length);
   if (thresholds.mode === ThresholdsMode.Percentage) {
     const { min, max } = getFieldConfigWithMinMax(field);
@@ -403,10 +282,10 @@ export function mergeThresholdValues(field: Field, theme: GrafanaTheme2): Field
       },
     },
     type: FieldType.string,
-    values: new ArrayVector(vals),
-    display: (value: string) => ({
-      text: value,
-      color: textToColor.get(value),
+    values: vals,
+    display: (value) => ({
+      text: String(value),
+      color: textToColor.get(String(value)),
       numeric: NaN,
     }),
   };
@@ -422,19 +301,68 @@ export function prepareTimelineFields(
   if (!series?.length) {
     return { warn: 'No data in response' };
   }
+
+  cacheFieldDisplayNames(series);
+
   let hasTimeseries = false;
   const frames: DataFrame[] = [];
 
   for (let frame of series) {
-    let isTimeseries = false;
+    let startFieldIdx = -1;
+    let endFieldIdx = -1;
+
+    for (let i = 0; i < frame.fields.length; i++) {
+      let f = frame.fields[i];
+
+      if (f.type === FieldType.time) {
+        if (startFieldIdx === -1) {
+          startFieldIdx = i;
+        } else if (endFieldIdx === -1) {
+          endFieldIdx = i;
+          break;
+        }
+      }
+    }
+
+    let isTimeseries = startFieldIdx !== -1;
     let changed = false;
-    let maybeSortedFrame = maybeSortFrame(
-      frame,
-      frame.fields.findIndex((f) => f.type === FieldType.time)
-    );
+    frame = maybeSortFrame(frame, startFieldIdx);
+
+    // if we have a second time field, assume it is state end timestamps
+    // and insert nulls into the data at the end timestamps
+    if (endFieldIdx !== -1) {
+      let startFrame: DataFrame = {
+        ...frame,
+        fields: frame.fields.filter((f, i) => i !== endFieldIdx),
+      };
+
+      let endFrame: DataFrame = {
+        length: frame.length,
+        fields: [frame.fields[endFieldIdx]],
+      };
+
+      frame = outerJoinDataFrames({
+        frames: [startFrame, endFrame],
+        keepDisplayNames: true,
+        nullMode: () => NULL_RETAIN,
+      })!;
+
+      frame.fields.forEach((f, i) => {
+        if (i > 0) {
+          let vals = f.values;
+          for (let i = 0; i < vals.length; i++) {
+            if (vals[i] == null) {
+              vals[i] = null;
+            }
+          }
+        }
+      });
+
+      changed = true;
+    }
 
     let nulledFrame = applyNullInsertThreshold({
-      frame: maybeSortedFrame,
+      frame,
       refFieldPseudoMin: timeRange.from.valueOf(),
       refFieldPseudoMax: timeRange.to.valueOf(),
     });
@@ -443,14 +371,20 @@ export function prepareTimelineFields(
       changed = true;
     }
 
+    frame = nullToValue(nulledFrame);
+
     const fields: Field[] = [];
-    for (let field of nullToValue(nulledFrame).fields) {
+    for (let field of frame.fields) {
+      if (field.config.custom?.hideFrom?.viz) {
+        continue;
+      }
       switch (field.type) {
         case FieldType.time:
           isTimeseries = true;
           hasTimeseries = true;
           fields.push(field);
           break;
+        case FieldType.enum:
         case FieldType.number:
           if (mergeValues && field.config.color?.mode === FieldColorModeId.Thresholds) {
             const f = mergeThresholdValues(field, theme);
@@ -473,6 +407,7 @@ export function prepareTimelineFields(
               },
             },
           };
+          changed = true;
           fields.push(field);
           break;
         default:
@@ -483,11 +418,11 @@ export function prepareTimelineFields(
       hasTimeseries = true;
       if (changed) {
         frames.push({
-          ...maybeSortedFrame,
+          ...frame,
           fields,
         });
       } else {
-        frames.push(maybeSortedFrame);
+        frames.push(frame);
       }
     }
   }
@@ -498,20 +433,46 @@ export function prepareTimelineFields(
   if (!frames.length) {
     return { warn: 'No graphable fields' };
   }
+
   return { frames };
 }
 
-export function getThresholdItems(fieldConfig: FieldConfig, theme: GrafanaTheme2): VizLegendItem[] {
+export function makeFramePerSeries(frames: DataFrame[]) {
+  const outFrames: DataFrame[] = [];
+
+  for (let frame of frames) {
+    const timeFields = frame.fields.filter((field) => field.type === FieldType.time);
+
+    if (timeFields.length > 0) {
+      for (let field of frame.fields) {
+        if (field.type !== FieldType.time) {
+          outFrames.push({ fields: [...timeFields, field], length: frame.length });
+        }
+      }
+    }
+  }
+
+  return outFrames;
+}
+
+export function getThresholdItems(
+  fieldConfig: FieldConfig,
+  theme: GrafanaTheme2,
+  thresholdItems?: ThresholdsConfig
+): VizLegendItem[] {
   const items: VizLegendItem[] = [];
-  const thresholds = fieldConfig.thresholds;
+  const thresholds = thresholdItems ? thresholdItems : fieldConfig.thresholds;
   if (!thresholds || !thresholds.steps.length) {
     return items;
   }
 
   const steps = thresholds.steps;
-  const disp = getValueFormat(thresholds.mode === ThresholdsMode.Percentage ? 'percent' : fieldConfig.unit ?? '');
+  const getDisplay = getValueFormat(
+    thresholds.mode === ThresholdsMode.Percentage ? 'percent' : (fieldConfig.unit ?? '')
+  );
 
-  const fmt = (v: number) => formattedValueToString(disp(v));
+  // `undefined` value for decimals will use `auto`
+  const format = (value: number) => formattedValueToString(getDisplay(value, fieldConfig.decimals ?? undefined));
 
   for (let i = 0; i < steps.length; i++) {
     let step = steps[i];
@@ -527,7 +488,7 @@ export function getThresholdItems(fieldConfig: FieldConfig, theme: GrafanaTheme2
     }
 
     items.push({
-      label: `${pre}${fmt(value)}${suf}`,
+      label: `${pre}${format(value)}${suf}`,
       color: theme.visualization.getColorByName(step.color),
       yAxis: 1,
     });
@@ -536,6 +497,66 @@ export function getThresholdItems(fieldConfig: FieldConfig, theme: GrafanaTheme2
   return items;
 }
 
+export function getValueMappingItems(mappings: ValueMapping[], theme: GrafanaTheme2): VizLegendItem[] {
+  const items: VizLegendItem[] = [];
+  if (!mappings) {
+    return items;
+  }
+
+  for (let mapping of mappings) {
+    const { options, type } = mapping;
+
+    if (type === MappingType.ValueToText) {
+      for (let [label, value] of Object.entries(options)) {
+        const color = value.color;
+        items.push({
+          label: label,
+          color: theme.visualization.getColorByName(color ?? FALLBACK_COLOR),
+          yAxis: 1,
+        });
+      }
+    }
+
+    if (type === MappingType.RangeToText) {
+      const { from, result, to } = options;
+      const { text, color } = result;
+      const label = text ? `[${from} - ${to}] ${text}` : `[${from} - ${to}]`;
+
+      items.push({
+        label: label,
+        color: theme.visualization.getColorByName(color ?? FALLBACK_COLOR),
+        yAxis: 1,
+      });
+    }
+
+    if (type === MappingType.RegexToText) {
+      const { pattern, result } = options;
+      const { text, color } = result;
+      const label = `${text || pattern}`;
+
+      items.push({
+        label: label,
+        color: theme.visualization.getColorByName(color ?? FALLBACK_COLOR),
+        yAxis: 1,
+      });
+    }
+
+    if (type === MappingType.SpecialValue) {
+      const { match, result } = options;
+      const { text, color } = result;
+      const label = `${text || match}`;
+
+      items.push({
+        label: label,
+        color: theme.visualization.getColorByName(color ?? FALLBACK_COLOR),
+        yAxis: 1,
+      });
+    }
+  }
+
+  return items;
+}
+
 export function prepareTimelineLegendItems(
   frames: DataFrame[] | undefined,
   options: VizLegendOptions,
@@ -559,6 +580,7 @@ export function getFieldLegendItem(fields: Field[], theme: GrafanaTheme2): VizLe
   const thresholds = fieldConfig.thresholds;
 
   // If thresholds are enabled show each step in the legend
+  // This ignores the hide from legend since the range is valid
   if (colorMode === FieldColorModeId.Thresholds && thresholds?.steps && thresholds.steps.length > 1) {
     return getThresholdItems(fieldConfig, theme);
   }
@@ -568,15 +590,17 @@ export function getFieldLegendItem(fields: Field[], theme: GrafanaTheme2): VizLe
     return undefined; // eventually a color bar
   }
 
-  let stateColors: Map<string, string | undefined> = new Map();
+  const stateColors: Map<string, string | undefined> = new Map();
 
   fields.forEach((field) => {
-    field.values.toArray().forEach((v) => {
-      let state = field.display!(v);
-      if (state.color) {
-        stateColors.set(state.text, state.color!);
-      }
-    });
+    if (!field.config.custom?.hideFrom?.legend) {
+      field.values.forEach((v) => {
+        let state = field.display!(v);
+        if (state.color) {
+          stateColors.set(state.text, state.color!);
+        }
+      });
+    }
   });
 
   stateColors.forEach((color, label) => {
@@ -612,13 +636,13 @@ export function findNextStateIndex(field: Field, datapointIdx: number) {
     return null;
   }
 
-  const startValue = field.values.get(datapointIdx);
+  const startValue = field.values[datapointIdx];
 
   while (end === undefined) {
     if (rightPointer >= field.values.length) {
       return null;
     }
-    const rightValue = field.values.get(rightPointer);
+    const rightValue = field.values[rightPointer];
 
     if (rightValue === undefined || rightValue === startValue) {
       rightPointer++;
@@ -674,19 +698,19 @@ export function fmtDuration(milliSeconds: number): string {
     yr > 0
       ? yr + 'y ' + (mo > 0 ? mo + 'mo ' : '') + (wk > 0 ? wk + 'w ' : '') + (d > 0 ? d + 'd ' : '')
       : mo > 0
-      ? mo + 'mo ' + (wk > 0 ? wk + 'w ' : '') + (d > 0 ? d + 'd ' : '')
-      : wk > 0
-      ? wk + 'w ' + (d > 0 ? d + 'd ' : '')
-      : d > 0
-      ? d + 'd ' + (h > 0 ? h + 'h ' : '')
-      : h > 0
-      ? h + 'h ' + (m > 0 ? m + 'm ' : '')
-      : m > 0
-      ? m + 'm ' + (s > 0 ? s + 's ' : '')
-      : s > 0
-      ? s + 's ' + (ms > 0 ? ms + 'ms ' : '')
-      : ms > 0
-      ? ms + 'ms '
-      : '0'
+        ? mo + 'mo ' + (wk > 0 ? wk + 'w ' : '') + (d > 0 ? d + 'd ' : '')
+        : wk > 0
+          ? wk + 'w ' + (d > 0 ? d + 'd ' : '')
+          : d > 0
+            ? d + 'd ' + (h > 0 ? h + 'h ' : '')
+            : h > 0
+              ? h + 'h ' + (m > 0 ? m + 'm ' : '')
+              : m > 0
+                ? m + 'm ' + (s > 0 ? s + 's ' : '')
+                : s > 0
+                  ? s + 's ' + (ms > 0 ? ms + 'ms ' : '')
+                  : ms > 0
+                    ? ms + 'ms '
+                    : '0'
   ).trim();
 }
