@@ -51,7 +51,7 @@ export class TemplateSrv implements BaseTemplateSrv {
   private regex = variableRegex;
   private index: any = {};
   private grafanaVariables = new Map<string, any>();
-  private timeRange?: TimeRange | null = null;
+  private _timeRange?: TimeRange | null = null;
   private _adhocFiltersDeprecationWarningLogged = new Map<string, boolean>();
 
   constructor(private dependencies: TemplateSrvDependencies = runtimeDependencies) {
@@ -60,7 +60,7 @@ export class TemplateSrv implements BaseTemplateSrv {
 
   init(variables: any, timeRange?: TimeRange) {
     this._variables = variables;
-    this.timeRange = timeRange;
+    this._timeRange = timeRange;
     this.updateIndex();
   }
 
@@ -83,6 +83,16 @@ export class TemplateSrv implements BaseTemplateSrv {
     return this.dependencies.getVariables();
   }
 
+  get timeRange(): TimeRange | null | undefined {
+    if (window.__grafanaSceneContext && window.__grafanaSceneContext.isActive) {
+      const sceneTimeRange = sceneGraph.getTimeRange(window.__grafanaSceneContext);
+
+      return sceneTimeRange.state.value;
+    }
+
+    return this._timeRange;
+  }
+
   updateIndex() {
     const existsOrEmpty = (value: unknown) => value || value === '';
 
@@ -110,7 +120,7 @@ export class TemplateSrv implements BaseTemplateSrv {
   }
 
   updateTimeRange(timeRange: TimeRange) {
-    this.timeRange = timeRange;
+    this._timeRange = timeRange;
     this.updateIndex();
   }
 
@@ -124,7 +134,7 @@ export class TemplateSrv implements BaseTemplateSrv {
    * interpolateVariablesInQueries or applyTemplateVariables it is passed as a new argument
    **/
   getAdhocFilters(datasourceName: string, skipDeprecationWarning?: boolean): AdHocVariableFilter[] {
-    let filters: any = [];
+    let filters: AdHocVariableFilter[] = [];
     let ds = getDataSourceSrv().getInstanceSettings(datasourceName);
 
     if (!ds) {
@@ -255,7 +265,8 @@ export class TemplateSrv implements BaseTemplateSrv {
         sceneObject,
         target,
         scopedVars,
-        format as string | VariableCustomFormatterFn | undefined
+        format as string | VariableCustomFormatterFn | undefined,
+        interpolations
       );
     }
 
@@ -265,7 +276,8 @@ export class TemplateSrv implements BaseTemplateSrv {
         window.__grafanaSceneContext,
         target,
         scopedVars,
-        format as string | VariableCustomFormatterFn | undefined
+        format as string | VariableCustomFormatterFn | undefined,
+        interpolations
       );
     }
 
